// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "CapsuleMesh.hpp"

using namespace Dawn;

void CapsuleMesh::calculateRing(
  const int32_t segments,
  const float_t height,
  const float_t radius,
  const float_t dr,
  const float_t y,
  const float_t dy,
  std::vector<glm::vec3> &positions
) {
  float_t segIncr = 1.0f / (float_t)(segments - 1);
  for(int32_t s = 0; s < segments; s++ ) {
    float_t x = cosf(MATH_PI * 2 * s * segIncr) * dr;
    float_t z = sinf(MATH_PI * 2 * s * segIncr) * dr;
    positions.emplace_back(glm::vec3(radius * x, radius * y + height * dy, radius * z ));
  }
}

void CapsuleMesh::create(
  Mesh &mesh,
  const float_t radius,
  const float_t height
) {
  std::vector<glm::vec3> positions;
  std::vector<meshindice_t> indices;

  const int32_t slices = 12;
  const int32_t segments = 12;
  const int32_t ringsBody = slices + 1;
  const int32_t ringsTotal = slices + ringsBody;

  positions.reserve(segments * ringsTotal);
  indices.reserve((segments - 1) * (ringsTotal - 1) * 6);

  const float_t bodyIncr = 1.0f / (float_t)(ringsBody - 1);
  const float_t ringIncr = 1.0f / (float_t)(slices - 1);
  for(int32_t r = 0; r < slices / 2; r++) {
    calculateRing(
      segments,
      height,
      radius,
      sinf(MATH_PI * r * ringIncr),
      sinf(MATH_PI * (r * ringIncr - 0.5f)),
      -0.5f,
      positions
    );
  }

  for(int32_t r = 0; r < ringsBody; r++ ) {
    calculateRing(
      segments,
      height,
      radius,
      1.0f,
      0.0f,
      r * bodyIncr - 0.5f,
      positions
    );
  }

  for(int32_t r = slices / 2; r < slices; r++) {
    calculateRing(
      segments,
      height,
      radius,
      sinf(MATH_PI * r * ringIncr),
      sinf(MATH_PI * (r * ringIncr - 0.5f)),
      0.5f,
      positions
    );
  }

  for(int32_t r = 0; r < ringsTotal - 1; r++ ) {
    for(int32_t s = 0; s < segments - 1; s++ ) {
      indices.push_back( (uint32_t)(r * segments + ( s + 1 )) );
      indices.push_back( (uint32_t)(r * segments + ( s + 0 )) );
      indices.push_back( (uint32_t)(( r + 1 ) * segments + ( s + 1 )) );

      indices.push_back( (uint32_t)(( r + 1 ) * segments + ( s + 0 )) );
      indices.push_back( (uint32_t)(( r + 1 ) * segments + ( s + 1 )) );
      indices.push_back( (uint32_t)(r * segments + s) );
    }
  }

  mesh.createBuffers(positions.size(), indices.size());
  mesh.bufferPositions(0, positions.data(), positions.size());
  mesh.bufferIndices(0, indices.data(), indices.size());
}